@@ -1,8 +1,13 @@
 //! Contains the common Wasmtime command line interface (CLI) flags.
 
-use anyhow::Result;
+use anyhow::{Context, Result};
 use clap::Parser;
-use std::time::Duration;
+use serde::Deserialize;
+use std::{
+    fmt, fs,
+    path::{Path, PathBuf},
+    time::Duration,
+};
 use wasmtime::Config;
 
 pub mod opt;
@@ -37,12 +42,17 @@ fn init_file_per_thread_logger(prefix: &'static str) {
 }
 
 wasmtime_option_group! {
-    #[derive(PartialEq, Clone)]
+    #[derive(PartialEq, Clone, Deserialize)]
+    #[serde(rename_all = "kebab-case", deny_unknown_fields)]
     pub struct OptimizeOptions {
         /// Optimization level of generated code (0-2, s; default: 2)
+        #[serde(default)]
+        #[serde(deserialize_with = "crate::opt::cli_parse_wrapper")]
         pub opt_level: Option<wasmtime::OptLevel>,
 
         /// Register allocator algorithm choice.
+        #[serde(default)]
+        #[serde(deserialize_with = "crate::opt::cli_parse_wrapper")]
         pub regalloc_algorithm: Option<wasmtime::RegallocAlgorithm>,
 
         /// Do not allow Wasm linear memories to move in the host process's
@@ -85,7 +95,9 @@ wasmtime_option_group! {
 
         /// Enable memory protection keys for the pooling allocator; this can
         /// optimize the size of memory slots.
-        pub pooling_memory_protection_keys: Option<bool>,
+        #[serde(default)]
+        #[serde(deserialize_with = "crate::opt::cli_parse_wrapper")]
+        pub pooling_memory_protection_keys: Option<wasmtime::MpkEnabled>,
 
         /// Sets an upper limit on how many memory protection keys (MPK) Wasmtime
         /// will use. (default: 16)
@@ -95,6 +107,10 @@ wasmtime_option_group! {
         /// copy-on-write mapping (default: yes)
         pub memory_init_cow: Option<bool>,
 
+        /// Threshold below which CoW images are guaranteed to be used and be
+        /// dense.
+        pub memory_guaranteed_dense_image_size: Option<u64>,
+
         /// The maximum number of WebAssembly instances which can be created
         /// with the pooling allocator.
         pub pooling_total_core_instances: Option<u32>,
@@ -131,10 +147,6 @@ wasmtime_option_group! {
         /// pooling allocator. (default: 100)
         pub pooling_max_unused_warm_slots: Option<u32>,
 
-        /// Configures whether or not stacks used for async futures are reset to
-        /// zero after usage. (default: false)
-        pub pooling_async_stack_zeroing: Option<bool>,
-
         /// How much memory, in bytes, to keep resident for async stacks allocated
         /// with the pooling allocator. (default: 0)
         pub pooling_async_stack_keep_resident: Option<usize>,
@@ -189,12 +201,15 @@ wasmtime_option_group! {
 }
 
 wasmtime_option_group! {
-    #[derive(PartialEq, Clone)]
+    #[derive(PartialEq, Clone, Deserialize)]
+    #[serde(rename_all = "kebab-case", deny_unknown_fields)]
     pub struct CodegenOptions {
         /// Either `cranelift` or `winch`.
         ///
         /// Currently only `cranelift` and `winch` are supported, but not all
         /// builds of Wasmtime have both built in.
+        #[serde(default)]
+        #[serde(deserialize_with = "crate::opt::cli_parse_wrapper")]
         pub compiler: Option<wasmtime::Strategy>,
         /// Which garbage collector to use: `drc` or `null`.
         ///
@@ -205,6 +220,8 @@ wasmtime_option_group! {
         ///
         /// Note that not all builds of Wasmtime will have support for garbage
         /// collection included.
+        #[serde(default)]
+        #[serde(deserialize_with = "crate::opt::cli_parse_wrapper")]
         pub collector: Option<wasmtime::Collector>,
         /// Enable Cranelift's internal debug verifier (expensive)
         pub cranelift_debug_verifier: Option<bool>,
@@ -221,6 +238,7 @@ wasmtime_option_group! {
         pub native_unwind_info: Option<bool>,
 
         #[prefixed = "cranelift"]
+        #[serde(default)]
         /// Set a cranelift-specific option. Use `wasmtime settings` to see
         /// all.
         pub cranelift: Vec<(String, Option<String>)>,
@@ -232,7 +250,8 @@ wasmtime_option_group! {
 }
 
 wasmtime_option_group! {
-    #[derive(PartialEq, Clone)]
+    #[derive(PartialEq, Clone, Deserialize)]
+    #[serde(rename_all = "kebab-case", deny_unknown_fields)]
     pub struct DebugOptions {
         /// Enable generation of DWARF debug information in compiled code.
         pub debug_info: Option<bool>,
@@ -252,7 +271,8 @@ wasmtime_option_group! {
 }
 
 wasmtime_option_group! {
-    #[derive(PartialEq, Clone)]
+    #[derive(PartialEq, Clone, Deserialize)]
+    #[serde(rename_all = "kebab-case", deny_unknown_fields)]
     pub struct WasmOptions {
         /// Enable canonicalization of all NaN values.
         pub nan_canonicalization: Option<bool>,
@@ -276,6 +296,9 @@ wasmtime_option_group! {
         /// difference between the two is how much stack the host has to execute
         /// on.
         pub async_stack_size: Option<usize>,
+        /// Configures whether or not stacks used for async futures are zeroed
+        /// before (re)use as a defense-in-depth mechanism. (default: false)
+        pub async_stack_zeroing: Option<bool>,
         /// Allow unknown exports when running commands.
         pub unknown_exports_allow: Option<bool>,
         /// Allow the main module to import unknown functions, using an
@@ -339,10 +362,8 @@ wasmtime_option_group! {
         pub memory64: Option<bool>,
         /// Configure support for the component-model proposal.
         pub component_model: Option<bool>,
-        /// Configure support for 33+ flags in the component model.
-        pub component_model_more_flags: Option<bool>,
-        /// Component model support for more than one return value.
-        pub component_model_multiple_returns: Option<bool>,
+        /// Component model support for async lifting/lowering.
+        pub component_model_async: Option<bool>,
         /// Configure support for the function-references proposal.
         pub function_references: Option<bool>,
         /// Configure support for the GC proposal.
@@ -361,7 +382,8 @@ wasmtime_option_group! {
 }
 
 wasmtime_option_group! {
-    #[derive(PartialEq, Clone)]
+    #[derive(PartialEq, Clone, Deserialize)]
+    #[serde(rename_all = "kebab-case", deny_unknown_fields)]
     pub struct WasiOptions {
         /// Enable support for WASI CLI APIs, including filesystems, sockets, clocks, and random.
         pub cli: Option<bool>,
@@ -390,6 +412,7 @@ wasmtime_option_group! {
         /// systemd listen fd specification (UNIX only)
         pub listenfd: Option<bool>,
         /// Grant access to the given TCP listen socket
+        #[serde(default)]
         pub tcplisten: Vec<String>,
         /// Implement WASI Preview1 using new Preview2 implementation (true, default) or legacy
         /// implementation (false)
@@ -402,6 +425,7 @@ wasmtime_option_group! {
         /// an OpenVINO model named `bar`. Note that which model encodings are
         /// available is dependent on the backends implemented in the
         /// `wasmtime_wasi_nn` crate.
+        #[serde(skip)]
         pub nn_graph: Vec<WasiNnGraph>,
         /// Flag for WASI preview2 to inherit the host's network within the
         /// guest so it has full access to all addresses/ports/etc.
@@ -421,8 +445,10 @@ wasmtime_option_group! {
         /// This option can be further overwritten with `--env` flags.
         pub inherit_env: Option<bool>,
         /// Pass a wasi config variable to the program.
+        #[serde(skip)]
         pub config_var: Vec<KeyValuePair>,
         /// Preset data for the In-Memory provider of WASI key-value API.
+        #[serde(skip)]
         pub keyvalue_in_memory_data: Vec<KeyValuePair>,
     }
 
@@ -444,7 +470,8 @@ pub struct KeyValuePair {
 }
 
 /// Common options for commands that translate WebAssembly modules
-#[derive(Parser, Clone)]
+#[derive(Parser, Clone, Deserialize)]
+#[serde(deny_unknown_fields)]
 pub struct CommonOptions {
     // These options groups are used to parse `-O` and such options but aren't
     // the raw form consumed by the CLI. Instead they're pushed into the `pub`
@@ -456,43 +483,70 @@ pub struct CommonOptions {
     /// Optimization and tuning related options for wasm performance, `-O help` to
     /// see all.
     #[arg(short = 'O', long = "optimize", value_name = "KEY[=VAL[,..]]")]
+    #[serde(skip)]
     opts_raw: Vec<opt::CommaSeparated<Optimize>>,
 
     /// Codegen-related configuration options, `-C help` to see all.
     #[arg(short = 'C', long = "codegen", value_name = "KEY[=VAL[,..]]")]
+    #[serde(skip)]
     codegen_raw: Vec<opt::CommaSeparated<Codegen>>,
 
     /// Debug-related configuration options, `-D help` to see all.
     #[arg(short = 'D', long = "debug", value_name = "KEY[=VAL[,..]]")]
+    #[serde(skip)]
     debug_raw: Vec<opt::CommaSeparated<Debug>>,
 
     /// Options for configuring semantic execution of WebAssembly, `-W help` to see
     /// all.
     #[arg(short = 'W', long = "wasm", value_name = "KEY[=VAL[,..]]")]
+    #[serde(skip)]
     wasm_raw: Vec<opt::CommaSeparated<Wasm>>,
 
     /// Options for configuring WASI and its proposals, `-S help` to see all.
     #[arg(short = 'S', long = "wasi", value_name = "KEY[=VAL[,..]]")]
+    #[serde(skip)]
     wasi_raw: Vec<opt::CommaSeparated<Wasi>>,
 
     // These fields are filled in by the `configure` method below via the
     // options parsed from the CLI above. This is what the CLI should use.
     #[arg(skip)]
+    #[serde(skip)]
     configured: bool,
+
     #[arg(skip)]
+    #[serde(rename = "optimize", default)]
     pub opts: OptimizeOptions,
+
     #[arg(skip)]
+    #[serde(rename = "codegen", default)]
     pub codegen: CodegenOptions,
+
     #[arg(skip)]
+    #[serde(rename = "debug", default)]
     pub debug: DebugOptions,
+
     #[arg(skip)]
+    #[serde(rename = "wasm", default)]
     pub wasm: WasmOptions,
+
     #[arg(skip)]
+    #[serde(rename = "wasi", default)]
     pub wasi: WasiOptions,
 
     /// The target triple; default is the host triple
     #[arg(long, value_name = "TARGET")]
+    #[serde(skip)]
     pub target: Option<String>,
+
+    /// Use the specified TOML configuration file.
+    /// This TOML configuration file can provide same configuration options as the
+    /// `--optimize`, `--codgen`, `--debug`, `--wasm`, `--wasi` CLI options, with a couple exceptions.
+    ///
+    /// Additional options specified on the command line will take precedent over options loaded from
+    /// this TOML file.
+    #[arg(long = "config", value_name = "FILE")]
+    #[serde(skip)]
+    pub config: Option<PathBuf>,
 }
 
 macro_rules! match_feature {
@@ -517,20 +571,48 @@ macro_rules! match_feature {
 }
 
 impl CommonOptions {
-    fn configure(&mut self) {
+    /// Creates a blank new set of [`CommonOptions`] that can be configured.
+    pub fn new() -> CommonOptions {
+        CommonOptions {
+            opts_raw: Vec::new(),
+            codegen_raw: Vec::new(),
+            debug_raw: Vec::new(),
+            wasm_raw: Vec::new(),
+            wasi_raw: Vec::new(),
+            configured: true,
+            opts: Default::default(),
+            codegen: Default::default(),
+            debug: Default::default(),
+            wasm: Default::default(),
+            wasi: Default::default(),
+            target: None,
+            config: None,
+        }
+    }
+
+    fn configure(&mut self) -> Result<()> {
         if self.configured {
-            return;
+            return Ok(());
         }
         self.configured = true;
+        if let Some(toml_config_path) = &self.config {
+            let toml_options = CommonOptions::from_file(toml_config_path)?;
+            self.opts = toml_options.opts;
+            self.codegen = toml_options.codegen;
+            self.debug = toml_options.debug;
+            self.wasm = toml_options.wasm;
+            self.wasi = toml_options.wasi;
+        }
         self.opts.configure_with(&self.opts_raw);
         self.codegen.configure_with(&self.codegen_raw);
         self.debug.configure_with(&self.debug_raw);
         self.wasm.configure_with(&self.wasm_raw);
         self.wasi.configure_with(&self.wasi_raw);
+        Ok(())
     }
 
     pub fn init_logging(&mut self) -> Result<()> {
-        self.configure();
+        self.configure()?;
         if self.debug.logging == Some(false) {
             return Ok(());
         }
@@ -563,7 +645,7 @@ impl CommonOptions {
     }
 
     pub fn config(&mut self, pooling_allocator_default: Option<bool>) -> Result<Config> {
-        self.configure();
+        self.configure()?;
         let mut config = Config::new();
 
         match_feature! {
@@ -661,21 +743,15 @@ impl CommonOptions {
             .opts
             .memory_reservation
             .or(self.opts.static_memory_maximum_size);
-        match_feature! {
-            ["signals-based-traps" : memory_reservation]
-            size => config.memory_reservation(size),
-            _ => err,
+        if let Some(size) = memory_reservation {
+            config.memory_reservation(size);
         }
 
-        match_feature! {
-            ["signals-based-traps" : self.opts.static_memory_forced]
-            enable => config.memory_may_move(!enable),
-            _ => err,
+        if let Some(enable) = self.opts.static_memory_forced {
+            config.memory_may_move(!enable);
         }
-        match_feature! {
-            ["signals-based-traps" : self.opts.memory_may_move]
-            enable => config.memory_may_move(enable),
-            _ => err,
+        if let Some(enable) = self.opts.memory_may_move {
+            config.memory_may_move(enable);
         }
 
         let memory_guard_size = self
@@ -683,25 +759,19 @@ impl CommonOptions {
             .static_memory_guard_size
             .or(self.opts.dynamic_memory_guard_size)
             .or(self.opts.memory_guard_size);
-        match_feature! {
-            ["signals-based-traps" : memory_guard_size]
-            size => config.memory_guard_size(size),
-            _ => err,
+        if let Some(size) = memory_guard_size {
+            config.memory_guard_size(size);
         }
 
         let mem_for_growth = self
             .opts
             .memory_reservation_for_growth
             .or(self.opts.dynamic_memory_reserved_for_growth);
-        match_feature! {
-            ["signals-based-traps" : mem_for_growth]
-            size => config.memory_reservation_for_growth(size),
-            _ => err,
+        if let Some(size) = mem_for_growth {
+            config.memory_reservation_for_growth(size);
         }
-        match_feature! {
-            ["signals-based-traps" : self.opts.guard_before_linear_memory]
-            enable => config.guard_before_linear_memory(enable),
-            _ => err,
+        if let Some(enable) = self.opts.guard_before_linear_memory {
+            config.guard_before_linear_memory(enable);
         }
         if let Some(enable) = self.opts.table_lazy_init {
             config.table_lazy_init(enable);
@@ -718,15 +788,14 @@ impl CommonOptions {
         if let Some(enable) = self.debug.address_map {
             config.generate_address_map(enable);
         }
-        match_feature! {
-            ["signals-based-traps" : self.opts.memory_init_cow]
-            enable => config.memory_init_cow(enable),
-            _ => err,
+        if let Some(enable) = self.opts.memory_init_cow {
+            config.memory_init_cow(enable);
         }
-        match_feature! {
-            ["signals-based-traps" : self.opts.signals_based_traps]
-            enable => config.signals_based_traps(enable),
-            _ => err,
+        if let Some(size) = self.opts.memory_guaranteed_dense_image_size {
+            config.memory_guaranteed_dense_image_size(size);
+        }
+        if let Some(enable) = self.opts.signals_based_traps {
+            config.signals_based_traps(enable);
         }
         if let Some(enable) = self.codegen.native_unwind_info {
             config.native_unwind_info(enable);
@@ -775,11 +844,6 @@ impl CommonOptions {
                     if let Some(max) = self.opts.pooling_max_unused_warm_slots {
                         cfg.max_unused_warm_slots(max);
                     }
-                    match_feature! {
-                        ["async" : self.opts.pooling_async_stack_zeroing]
-                        enable => cfg.async_stack_zeroing(enable),
-                        _ => err,
-                    }
                     match_feature! {
                         ["async" : self.opts.pooling_async_stack_keep_resident]
                         size => cfg.async_stack_keep_resident(size),
@@ -805,11 +869,7 @@ impl CommonOptions {
                     }
                     match_feature! {
                         ["memory-protection-keys" : self.opts.pooling_memory_protection_keys]
-                        enable => cfg.memory_protection_keys(if enable {
-                            wasmtime::MpkEnabled::Enable
-                        } else {
-                            wasmtime::MpkEnabled::Disable
-                        }),
+                        enable => cfg.memory_protection_keys(enable),
                         _ => err,
                     }
                     match_feature! {
@@ -828,14 +888,14 @@ impl CommonOptions {
             true => err,
         }
 
-        if self.opts.pooling_memory_protection_keys.unwrap_or(false)
+        if self.opts.pooling_memory_protection_keys.is_some()
             && !self.opts.pooling_allocator.unwrap_or(false)
         {
             anyhow::bail!("memory protection keys require the pooling allocator");
         }
 
         if self.opts.pooling_max_memory_protection_keys.is_some()
-            && !self.opts.pooling_memory_protection_keys.unwrap_or(false)
+            && !self.opts.pooling_memory_protection_keys.is_some()
         {
             anyhow::bail!(
                 "max memory protection keys requires memory protection keys to be enabled"
@@ -847,6 +907,11 @@ impl CommonOptions {
             size => config.async_stack_size(size),
             _ => err,
         }
+        match_feature! {
+            ["async" : self.wasm.async_stack_zeroing]
+            enable => config.async_stack_zeroing(enable),
+            _ => err,
+        }
 
         if let Some(max) = self.wasm.max_wasm_stack {
             config.max_wasm_stack(max);
@@ -922,8 +987,7 @@ impl CommonOptions {
 
         handle_conditionally_compiled! {
             ("component-model", component_model, wasm_component_model)
-            ("component-model", component_model_more_flags, wasm_component_model_more_flags)
-            ("component-model", component_model_multiple_returns, wasm_component_model_multiple_returns)
+            ("component-model-async", component_model_async, wasm_component_model_async)
             ("threads", threads, wasm_threads)
             ("gc", gc, wasm_gc)
             ("gc", reference_types, wasm_reference_types)
@@ -931,4 +995,209 @@ impl CommonOptions {
         }
         Ok(())
     }
+
+    pub fn from_file<P: AsRef<Path>>(path: P) -> Result<Self> {
+        let path_ref = path.as_ref();
+        let file_contents = fs::read_to_string(path_ref)
+            .with_context(|| format!("failed to read config file: {path_ref:?}"))?;
+        toml::from_str::<CommonOptions>(&file_contents)
+            .with_context(|| format!("failed to parse TOML config file {path_ref:?}"))
+    }
+}
+
+#[cfg(test)]
+mod tests {
+    use wasmtime::{OptLevel, RegallocAlgorithm};
+
+    use super::*;
+
+    #[test]
+    fn from_toml() {
+        // empty toml
+        let empty_toml = "";
+        let mut common_options: CommonOptions = toml::from_str(empty_toml).unwrap();
+        common_options.config(None).unwrap();
+
+        // basic toml
+        let basic_toml = r#"
+            [optimize]
+            [codegen]
+            [debug]
+            [wasm]
+            [wasi]
+        "#;
+        let mut common_options: CommonOptions = toml::from_str(basic_toml).unwrap();
+        common_options.config(None).unwrap();
+
+        // toml with custom deserialization to match CLI flag parsing
+        for (opt_value, expected) in [
+            ("0", Some(OptLevel::None)),
+            ("1", Some(OptLevel::Speed)),
+            ("2", Some(OptLevel::Speed)),
+            ("\"s\"", Some(OptLevel::SpeedAndSize)),
+            ("\"hello\"", None), // should fail
+            ("3", None),         // should fail
+        ] {
+            let toml = format!(
+                r#"
+                    [optimize]
+                    opt-level = {opt_value}
+                "#,
+            );
+            let parsed_opt_level = toml::from_str::<CommonOptions>(&toml)
+                .ok()
+                .and_then(|common_options| common_options.opts.opt_level);
+
+            assert_eq!(
+                parsed_opt_level, expected,
+                "Mismatch for input '{opt_value}'. Parsed: {parsed_opt_level:?}, Expected: {expected:?}"
+            );
+        }
+
+        // Regalloc algorithm
+        for (regalloc_value, expected) in [
+            ("\"backtracking\"", Some(RegallocAlgorithm::Backtracking)),
+            ("\"single-pass\"", Some(RegallocAlgorithm::SinglePass)),
+            ("\"hello\"", None), // should fail
+            ("3", None),         // should fail
+            ("true", None),      // should fail
+        ] {
+            let toml = format!(
+                r#"
+                    [optimize]
+                    regalloc-algorithm = {regalloc_value}
+                "#,
+            );
+            let parsed_regalloc_algorithm = toml::from_str::<CommonOptions>(&toml)
+                .ok()
+                .and_then(|common_options| common_options.opts.regalloc_algorithm);
+            assert_eq!(
+                parsed_regalloc_algorithm, expected,
+                "Mismatch for input '{regalloc_value}'. Parsed: {parsed_regalloc_algorithm:?}, Expected: {expected:?}"
+            );
+        }
+
+        // Strategy
+        for (strategy_value, expected) in [
+            ("\"cranelift\"", Some(wasmtime::Strategy::Cranelift)),
+            ("\"winch\"", Some(wasmtime::Strategy::Winch)),
+            ("\"hello\"", None), // should fail
+            ("5", None),         // should fail
+            ("true", None),      // should fail
+        ] {
+            let toml = format!(
+                r#"
+                    [codegen]
+                    compiler = {strategy_value}
+                "#,
+            );
+            let parsed_strategy = toml::from_str::<CommonOptions>(&toml)
+                .ok()
+                .and_then(|common_options| common_options.codegen.compiler);
+            assert_eq!(
+                parsed_strategy, expected,
+                "Mismatch for input '{strategy_value}'. Parsed: {parsed_strategy:?}, Expected: {expected:?}",
+            );
+        }
+
+        // Collector
+        for (collector_value, expected) in [
+            (
+                "\"drc\"",
+                Some(wasmtime::Collector::DeferredReferenceCounting),
+            ),
+            ("\"null\"", Some(wasmtime::Collector::Null)),
+            ("\"hello\"", None), // should fail
+            ("5", None),         // should fail
+            ("true", None),      // should fail
+        ] {
+            let toml = format!(
+                r#"
+                    [codegen]
+                    collector = {collector_value}
+                "#,
+            );
+            let parsed_collector = toml::from_str::<CommonOptions>(&toml)
+                .ok()
+                .and_then(|common_options| common_options.codegen.collector);
+            assert_eq!(
+                parsed_collector, expected,
+                "Mismatch for input '{collector_value}'. Parsed: {parsed_collector:?}, Expected: {expected:?}",
+            );
+        }
+    }
+}
+
+impl Default for CommonOptions {
+    fn default() -> CommonOptions {
+        CommonOptions::new()
+    }
+}
+
+impl fmt::Display for CommonOptions {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let CommonOptions {
+            codegen_raw,
+            codegen,
+            debug_raw,
+            debug,
+            opts_raw,
+            opts,
+            wasm_raw,
+            wasm,
+            wasi_raw,
+            wasi,
+            configured,
+            target,
+            config,
+        } = self;
+        if let Some(target) = target {
+            write!(f, "--target {target} ")?;
+        }
+        if let Some(config) = config {
+            write!(f, "--config {} ", config.display())?;
+        }
+
+        let codegen_flags;
+        let opts_flags;
+        let wasi_flags;
+        let wasm_flags;
+        let debug_flags;
+
+        if *configured {
+            codegen_flags = codegen.to_options();
+            debug_flags = debug.to_options();
+            wasi_flags = wasi.to_options();
+            wasm_flags = wasm.to_options();
+            opts_flags = opts.to_options();
+        } else {
+            codegen_flags = codegen_raw
+                .iter()
+                .flat_map(|t| t.0.iter())
+                .cloned()
+                .collect();
+            debug_flags = debug_raw.iter().flat_map(|t| t.0.iter()).cloned().collect();
+            wasi_flags = wasi_raw.iter().flat_map(|t| t.0.iter()).cloned().collect();
+            wasm_flags = wasm_raw.iter().flat_map(|t| t.0.iter()).cloned().collect();
+            opts_flags = opts_raw.iter().flat_map(|t| t.0.iter()).cloned().collect();
+        }
+
+        for flag in codegen_flags {
+            write!(f, "-C{flag} ")?;
+        }
+        for flag in opts_flags {
+            write!(f, "-O{flag} ")?;
+        }
+        for flag in wasi_flags {
+            write!(f, "-S{flag} ")?;
+        }
+        for flag in wasm_flags {
+            write!(f, "-W{flag} ")?;
+        }
+        for flag in debug_flags {
+            write!(f, "-D{flag} ")?;
+        }
+
+        Ok(())
+    }
 }
